#--------------------------------------------------------------------------------------------------------------------------
# Read Values from Solis via Modbus
#
# This module reades the binary via modbus and generate a dictionary
#
# License:
# Feel free to use this code in your private or business environment or as example for your own code.
# Redistribution is only allowed with no changes.
# For any other kind of usage, please ask me.
# Ralf Amandi, Avarix GmbH
#
# 19.04.2021 Version 1.0
#--------------------------------------------------------------------------------------------------------------------------
 
import time
import serial
import minimalmodbus
from datetime import datetime

instrument = None
hexchars=['0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','D']
workmodeText={'0':'No response mode','1':'Volt–watt default','2':'Volt–var','3':'Fixed power factor ','4':'Fix reactive power','5':'Power-PF','6':'Rule21Volt–watt'}
statusMessage={"0000":"Waiting", "0001":"OpenRun", "0002":"SoftRun", "0003":"Generating", "1004":"GridOff" }
rs485Config = {"Dev":"/dev/ttyUSB0","Addr":"1"} # see cfg Dev

#--------------------------------------------------------------------------------------------------------------------------
# Setup from Config
#--------------------------------------------------------------------------------------------------------------------------

def SetConfig(config):
	global rs485Config, workmodeText, statusMessage
	rs485Config=config["RS485"]
	workmodeText=config["Workmode"]
	statusMessage=config["Statusmessage"]


#--------------------------------------------------------------------------------------------------------------------------
# Modbus
#--------------------------------------------------------------------------------------------------------------------------
def InitModBus():
	global instrument,rs485Config
	print (rs485Config["Addr"])
	instrument = minimalmodbus.Instrument(rs485Config["Dev"], int(rs485Config["Addr"])) # port name, slave address 
	instrument.serial.baudrate = 9600   # Baud
	instrument.serial.bytesize = 8
	instrument.serial.parity   = serial.PARITY_NONE
	instrument.serial.stopbits = 1
	instrument.serial.timeout  = 0.5   # seconds

def CloseModBus():
	global instrument,rs485Config
	instrument = None
	
def ReadMulti(regno,cnt):
	global instrument
	try:
		v=instrument.read_registers(regno,  functioncode=4, number_of_registers=cnt)
	except BaseException as e:
		time.sleep(1);
		v=instrument.read_registers(regno,  functioncode=4, number_of_registers=cnt)
	#print("D:",json.dumps(v));
	return v;

def ToS16(a):
	if (a>>15 ==1):
		return a-65536;
	else:
		return a;
	

def ToS32(h,l):
	if (h>>15 ==1):
		return h*65536+l-4294967296;
	else:
		return h*65536+l;

def Fmt2(val):
	if (val<10):
		return "0"+str(val)
	else:
		return str(val)

def HexW(v):
	return hexchars[ (v>>12) & 15] + hexchars[ (v>>8) & 15] +  hexchars[ (v>>4) & 15] + hexchars[ v & 15]

def Hex_SN(v): # für Seriennummer
	return hexchars[ v & 15] + hexchars[ (v>>4) & 15] + hexchars[ (v>>8) & 15] + hexchars[ (v>>12) & 15]
	
   
#--------------------------------------------------------------------------------------------------------------------------
# Interpolate the values to get higher precision
#--------------------------------------------------------------------------------------------------------------------------
# The Summary values have a very bad precision
# - Every Time a new values arrives, the function "syncs" to that value
# - The calculated values is alway in the interval LASTVAL..LASTVAL+PRECISION
interpTotal = {"LASTVAL":0, "CURRVAL":0, "LASTSTAMP":0, "LASTPOW":0, "PRECISION":1}
interpYear =  {"LASTVAL":0, "CURRVAL":0, "LASTSTAMP":0, "LASTPOW":0, "PRECISION":1}
interpMonth = {"LASTVAL":0, "CURRVAL":0, "LASTSTAMP":0, "LASTPOW":0, "PRECISION":1}
interpToday = {"LASTVAL":0, "CURRVAL":0, "LASTSTAMP":0, "LASTPOW":0, "PRECISION":0.1}

def interpolate(data, val,ts,pow):
	if (val != data["LASTVAL"]):
		data["LASTVAL"] = val
		print("interpolate Last:",data["LASTVAL"], "Curr: ",data["CURRVAL"], "Now", val);
	else:
		t = ts -data["LASTSTAMP"]
		val = round(data["CURRVAL"] + (pow+data["LASTPOW"])/2000 *t/3600,3)
		if val>data["LASTVAL"]+ data["PRECISION"]: val = data["LASTVAL"]+ data["PRECISION"]
		print("interpolate Last:",data["LASTVAL"], "Curr: ",data["CURRVAL"], "Pw", pow, "t",t);		
	data["CURRVAL"] = val;
	data["LASTSTAMP"] = ts;
	data["LASTPOW"] = pow;
	return val

#--------------------------------------------------------------------------------------------------------------------------
# Read from RS485 to Hash Array
#--------------------------------------------------------------------------------------------------------------------------

def DecodeStatus(code): # for 4g and other, not for 1P 2G
	try:
		return statusMessage[code]
	except:
		return "Alarm "+code
	

def ReadSolisValues():
	v = {}
	energy={}
	ac = {"VOLT":[], "CURR":[], "POW":[]}
	mpp = {"VOLT":[], "CURR":[], "POW":[]}
	dc = {}
	limit = {}
	sysinfo = {}
	status = {}
	config = {}
	#-- Serial / Datetime
	r = ReadMulti(3061-1,4)
	sysinfo['SERIAL']=(Hex_SN(r[0])+Hex_SN(r[1])+Hex_SN(r[2])+Hex_SN(r[3]))[0:15]
	r = ReadMulti(3073-1,6)
	dt = datetime (2000+r[0],r[1],r[2],r[3],r[4],r[5])
	dt_stamp = dt.timestamp()
	status['TIMESTAMP']= dt_stamp
	status['DATETIME'] = dt.isoformat(sep=' ', timespec='seconds')
	
	#--------------------
	r = ReadMulti(3000-1,30)
	sysinfo['Model']=hex(r[0])
	sysinfo['DSP_Version']=hex(r[1])
	sysinfo['LCD_Version']=hex(r[2])
	sysinfo['HMI_Version']=r[21]
	ac['Type']=r[3]
	dc['Type']=r[4]
	ac['POWER']=ToS32(r[5],r[6])
	mpp['POWER']=ToS32(r[7],r[8])
	#-- Energy --
	energy['TOTAL']=interpolate(interpTotal,r[9]*65536+r[10],dt_stamp,ac['POWER']) # KWh
	energy['MONTH']=interpolate(interpMonth,r[11]*65536+r[12],dt_stamp,ac['POWER']) # KWh
	energy['LASTMONTH']=r[13]*65536+r[14] # KW
	energy['TODAY']=interpolate(interpToday,r[15] / 10,dt_stamp,ac['POWER']) # 0.1KWh
	energy['YESTERDAY']=r[16] / 10 # KWh
	energy['YEAR']=interpolate(interpYear,r[17]*65536+r[18],dt_stamp,ac['POWER']) # KWh
	energy['LASTYEAR']=r[19]*65536+r[20] # KWh
	#-- MPP --
	mpp['COUNT']=r[4]+1
	for i in range(mpp['COUNT']):
		mpp['VOLT'].append(r[22+i*2] / 10) # Volt
		mpp['CURR'].append(r[23+i*2] / 10) # Amp
		mpp['POW'].append(r[22+i*2] * r[23+i*2] / 100) # W
	#--
	r = ReadMulti(3030-1,30)
	status['ALARM_CODE'] = r[0]
	dc['INIT_GROUND_VOLT'] = r[1] / 10
	dc['BUSBAR_VOLT'] = r[2] / 10
	dc['HALFBUSBAR_VOLT'] = r[3] / 10
	#-- AC --
	if (ac['Type'] == 0):
		ac["COUNT"]=1
		ac['VOLT'].append(r[8] / 10)
		ac['CURR'].append(r[9] / 10)
		ac['POW'].append(r[8]*r[9] / 100)
	else: 
		ac["COUNT"]=3
		for i in range(3):
			ac['VOLT'].append(r[4+i] / 10) # Phase or Line voltage
			ac['CURR'].append(r[7+i] / 10) # Phase or Line current
			ac['POW'].append(r[4+i] * r[7+i] / 100) # Power
	#---
	config['DSP_MASTER_SLAVE_UPG_FLG'] = r[10]
	config['WORKMODE']={"CODE":Fmt2(r[11]),"TEXT":workmodeText[str(r[11])]}
	status['INV_TEMP'] = r[12] / 10
	ac["FREQ"]=r[13] / 100
	status['INV_STATUS'] = {"CODE":Fmt2(r[14]),"TEXT":DecodeStatus(HexW(r[14]))}
	#-- LIMIT --
	limit['ACTIVE_POWADJ']= ToS32(r[15],r[16]) # W Limit active power adjustment rated power
	limit['REACTIVE_POWADJ']= ToS32(r[17],r[18]) # VAR Limit active power adjustment rated power
	#config['INVERTER_CONTROL_WORD']=r[19]
	limit['ACTIVE'] = r[20] /100
	limit['ACT_ADJVAL']=ToS16(r[21]) / 1000  # Writeable; Actual adjust value of power factor
	limit['ACT_ADJVAL_FIX']=ToS16(r[22]) / 1000  # % Writeable;  Actual adjust value of power factor; for Fixed Mode (03)
	limit['REACTIVE_POW']=ToS16(r[23]) / 100  # % Writeable; Limit Reactive power; for 04 Fix reactive power (Blindleistung) 
	#---
	config['COUNTRY'] = r[24] # ? 14=EN50438NL (for 4G)
	config['POWER_CURVE_CODE'] = r[25]
	status['REACTIVE_POWER'] = ToS32(r[26],r[27]) # VAR; Blindleistung
	status['ACTIVE_POWER'] = ToS32(r[28],r[29]) # W Wirkleistung
	# 3060 (r[30]) not used
	# 3061ff Serial, 3073ff Datetime
	#-----------------------------
	r = ReadMulti(3066-1,7);
	status['Flags']={'Setting':r[0],"FaultCode":[r[1],r[2],r[3],r[4],r[5]],"WorkStatus":r[6]};
	v['SYSINFO']=sysinfo
	v['STATUS']=status
	v['ENERGY']=energy
	v["AC"]=ac
	v["DC"]=dc
	v['MPP']=mpp
	v['LIMIT']=limit
	v['CONFIG']=config
	return v

