#!/usr/bin/python3
#--------------------------------------------------------------------------------------------------------------------------
# Read Values from Solis via Modbus
#
# This tool
# - reads Solis PV-Inverter values to an dictionary
# - append as json to a log  
# - upsert to an MySql database
# - run an external program
#
# License:
# Feel free to use this code in your private or business environment or as example for your own code.
# Redistribution is only allowed with no changes.
# For any other kind of usage, please ask me.
# Ralf Amandi, Avarix GmbH
#
# 05.05.2021 Version 1.1
#--------------------------------------------------------------------------------------------------------------------------
import time
import os
import json
import ReadFromRS485
import SqlUpsert
import configparser
import minimalmodbus
from datetime import datetime
import logger
import small_server

#--------------------------------------------------------------------------------------------------------------------------
# Variables & Config
#--------------------------------------------------------------------------------------------------------------------------
testOnly = False	# Only print, no upsert

cfgFile = __file__[0:len(__file__)-3]+".cfg"
config = configparser.ConfigParser()
config.read(cfgFile)

logValuesFile = config["GENERAL"]["LogValuesFile"]
logValuesLevel = int (config["GENERAL"]["LogValuesLevel"])

exeFile=config["GENERAL"]["ExecuteFile"]
exeParam = config["GENERAL"]["ExecuteParam"]
pollIntervall=config["GENERAL"]["PollInterval"]
logInterval=config["GENERAL"]["LogInterval"]
lastLogTime=0

SqlUpsert.SetConfig(config)
ReadFromRS485.SetConfig(config)
logger.SetConfig(config)
small_server.SetConfig(config)

#--------------------------------------------------------------------------------------------------------------------------
# Subs for main loop
#--------------------------------------------------------------------------------------------------------------------------
def ReadAndProcessData():
	global exeFile,exeParam, logValuesFile,config, lastLogTime
	vals = ReadFromRS485.ReadSolisValues()
	logger.logMessage(3,"------ "+ vals["STATUS"]["DATETIME"]+" "+vals["SYSINFO"]["SERIAL"]+" ------\n")
	logger.logMessage(3,json.dumps(vals))
	small_server.last_value=vals

	# Log?
	if (time.time()>=lastLogTime+int(logInterval)):
		lastLogTime=time.time()
		if (logValuesLevel >=1 ): logger.appendFile(logValuesFile,json.dumps(vals)+"\n");
		SqlUpsert.Insert(vals)		
		if (exeFile !=""):
			os.system(exeFile+" "+str(eval(exeParam,{"v":vals,"json":json.dumps(vals),"config":config})))

    
#--------------------------------------------------------------------------------------------------------------------------
# Program
#--------------------------------------------------------------------------------------------------------------------------
online = False
logger.logMessage(2,"========== RESTART "+time.asctime( time.localtime(time.time()) )+" ==========")
try:
	ReadFromRS485.InitModBus()

	if (testOnly): # onl read & print
		vals = ReadFromRS485.ReadSolisValues()
		print (json.dumps(vals))
		sql = SqlUpsert.GenerateInsertSQL(vals)
		print(sql)
	elif (pollIntervall != ""):
		small_server.start_treating_server()
		pollIntervall = int(pollIntervall)
		while 1:
			try:
				print ("READ\n")
				ReadAndProcessData()
				if not online: 
					logger.logMessage(1,"Inverter is online");
					online = True
			except BaseException as e:
				logger.logMessage(1,"Exception in main: "+str(e))
				small_server.last_value={"STATUS":{"INV_STATUS":{"CODE":"0000","TEXT":"Waiting (offline)"}}}
	#		except minimalmodbus.NoResponseError as e:
				print ("No Response ",e)
				if online: 
					logger.logMessage(1,"Inverter seems to be offline");
					online = False
			ReadFromRS485.CloseModBus()
			t = pollIntervall - (time.time() % pollIntervall)
			#print ("SLEEP",t,"\n")
			time.sleep(t)
			#print ("SLEEP2\n")
			ReadFromRS485.InitModBus()
	else: # once
		ReadAndProcessData()
		
except BaseException as e:
	logger.logMessage(1,"Exception in main: "+str(e))
logger.logMessage(2,"=== Terminating ===")
